from .deconvolution import (
    simulate_elispot_counts,
    create_linear_system,
    solve_linear_system,
    SpotCounts,
    DeconvolutionResult,
)
from .initialization import init
from .optimization import optimize
from .solution import Solution
from .validity import is_valid, count_violations, violations_per_replicate
from .main import find_best_solution

__version__ = "1.9.3"

__all__ = [
    "__version__",
    "find_best_solution",
    "Solution",
    "init",
    "optimize",
    "count_violations",
    "is_valid",
    "violations_per_replicate",
    "simulate_elispot_counts",
    "create_linear_system",
    "solve_linear_system",
    "SpotCounts",
    "DeconvolutionResult",
]



from typing import Iterable, Mapping

Replicate = int
Peptide = int
Pool = int
SwapCandidateList = Iterable[tuple[Replicate, Pool, Peptide]]
ReplicateToNeighborDict = Mapping[Replicate, Mapping[Peptide, set[Peptide]]]
PeptidePairList = Iterable[tuple[Peptide, Peptide]]



import math
from typing import Optional, Literal

from .initialization import init
from .optimization import optimize
from .solution import Solution
from .types import Replicate, Pool, Peptide, PeptidePairList
from .validity import count_violations


def find_best_solution(
    num_peptides: int = 100,
    max_peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools_per_replicate: Optional[int | dict[Replicate, int]] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    """
    Try several different initialization methods and return the solution which
    is optimized to have the fewest violations and the fewest pools.

    Args
    ----
    num_peptides
        The number total number of peptides in the experiment

    max_peptides_per_pool
        Maximum number of peptides which can be in any one pool. This is also the number of peptides
        which will be in most pools if allow_extra_pools is False

    num_replicates
        Number of replicates in the experiment (i.e. how many distinct pools each peptide occurs in)

    num_pools_per_replicate
        Number of pools in each replicate. If None, then this will be set to the minimum number of pools
        required to fit all peptides. If an int, then this will be the number of pools in each replicate.

    invalid_neighbors
        List of peptide pairs which cannot be in the same pool

    preferred_neighbors
        List of peptide pairs which should be in the same pool if possible

    allow_extra_pools
        If True, then the solution can have more than the minimum number of pools required to fit all peptides. If False,
        then the returned solution may not be valid (i.e. some peptide pairs will occur in more than one pool together)

    verbose
        If True, then print out information about the solution as it is being constructed
    """
    shared_kwargs = dict(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        num_pools_per_replicate=num_pools_per_replicate,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        allow_extra_pools=allow_extra_pools,
        verbose=verbose,
    )

    if len(preferred_neighbors) > 0:
        if allow_extra_pools:
            # these are the two methods which can group
            # preferred peptides in pools, either during
            # initialization or during merging
            init_strategies = ["greedy", "singleton"]
        else:
            # if we don't allow extra pools, then we can only
            # group preferred peptides during initialization
            init_strategies = ["greedy"]
    elif allow_extra_pools:
        # if there are no preferred peptides, then we can
        # only use all the initialization methods
        init_strategies = ["greedy", "random", "valid", "singleton"]
    else:
        # only greedy and random initialization methods
        # let us be strict about the number of pools
        init_strategies = ["greedy", "random"]

    solutions = {
        strategy: init(strategy=strategy, **shared_kwargs)
        for strategy in init_strategies
    }
    best_solution = None
    best_violations = None
    best_num_pools = None

    for strategy, s in solutions.items():
        print(
            "Initialized with strategy '%s': violations=%d, num_pools=%d"
            % (strategy, count_violations(s), s.num_pools())
        )
        optimize(s, allow_extra_pools=allow_extra_pools, verbose=verbose)
        violations = count_violations(s)
        num_pools = s.num_pools()
        print(
            "-- after optimization of '%s' solution: violations=%d, num_pools=%d"
            % (strategy, violations, num_pools)
        )
        if (
            best_solution is None
            or (violations < best_violations)
            or (violations == best_violations and num_pools < best_num_pools)
        ):
            best_solution = s
            best_violations = violations
            best_num_pools = num_pools
            print("^^ new best solution")
    return best_solution



from typing import Mapping
from dataclasses import dataclass

import numpy as np

from .types import Replicate, Pool, Peptide
from .solution import Solution

SpotCounts = Mapping[Replicate, Mapping[Pool, int]]


def simulate_elispot_counts(
    s: Solution,
    num_hits=5,
    max_activity_per_well=2000,
    min_background_peptide_activity=0,
    max_background_peptide_activity=20,
    min_hit_peptide_activity=50,
    max_hit_peptide_activity=500,
    verbose=False,
) -> tuple[SpotCounts, set[Peptide]]:
    num_peptides = s.num_peptides

    background_peptide_activity_range = (
        max_background_peptide_activity - min_background_peptide_activity
    )
    hit_peptide_activity_range = max_hit_peptide_activity - min_hit_peptide_activity

    all_peptides = np.arange(num_peptides)
    np.random.shuffle(all_peptides)
    hit_peptides = all_peptides[:num_hits]
    not_hit_peptides = all_peptides[num_hits:]
    if verbose:
        print("Hits: %s" % (hit_peptides,))

    background = (
        np.random.rand(num_peptides) * background_peptide_activity_range
        + min_background_peptide_activity
    )
    if verbose:
        print("Background activity: %s" % (background,))

    hit_activity = (
        np.random.rand(num_peptides) * hit_peptide_activity_range
        + min_hit_peptide_activity
    )
    hit_activity[not_hit_peptides] = 0
    if verbose:
        print("Hit activity: %s" % (hit_activity,))

    spot_counts: SpotCounts = {
        r: {
            pool: min(
                max_activity_per_well,
                int(sum([background[i] + hit_activity[i] for i in peptides])),
            )
            for (pool, peptides) in d.items()
        }
        for (r, d) in s.assignments.items()
    }
    if verbose:
        print("Spot counts: %s" % (spot_counts,))
    return (spot_counts, set(hit_peptides))


@dataclass
class LinearSystem:
    A: np.ndarray
    b: np.ndarray
    pool_tuple_to_idx: Mapping[tuple[Replicate, Pool], int]
    idx_to_pool_tuple: Mapping[int, tuple[Replicate, Pool]]


def create_linear_system(
    s: Solution, spot_counts: SpotCounts, verbose=False
) -> LinearSystem:
    num_peptides = s.num_peptides
    num_pools = s.num_pools()

    A = np.zeros((num_pools, num_peptides + 1)).astype(float)
    b = np.zeros(num_pools).astype(float)

    pool_tuple_to_idx = {}
    idx_to_pool_tuple = {}
    i = 0
    for r, d in spot_counts.items():
        for pool, spots in d.items():
            b[i] = spots
            pool_tuple_to_idx[(r, pool)] = i
            idx_to_pool_tuple[i] = (r, pool)
            for p in s.assignments[r][pool]:
                A[i, p] = 1
            # add a ones column for a constant offset
            A[i, num_peptides] = 1
            i += 1
    if verbose:
        print("Ax = b")
        print("=======")
        print("A.shape: %s" % (A.shape,))
        print("b.shape: %s" % (b.shape,))
        print("A:\n%s" % (A,))
        print("A col sums: %s" % (A.sum(axis=0)))
        print("A row sums: %s" % (A.sum(axis=1)))
        print("b:\n%s" % (b,))
    return LinearSystem(
        A=A,
        b=b,
        pool_tuple_to_idx=pool_tuple_to_idx,
        idx_to_pool_tuple=idx_to_pool_tuple,
    )


@dataclass
class DeconvolutionResult:
    activity_per_peptide: np.ndarray
    prob_hit_per_peptide: np.ndarray
    high_confidence_hits: set[Peptide]


def solve_linear_system(
    linear_system: LinearSystem,
    min_peptide_activity: float = 1.0,
    leave_on_out=True,
    sparse_solution=True,
    verbose=False,
) -> DeconvolutionResult:
    from sklearn.linear_model import Lasso, Ridge

    A = linear_system.A
    b = linear_system.b
    num_pools, num_peptides_with_constant = A.shape
    num_peptides = num_peptides_with_constant - 1
    row_indices = list(range(num_pools))
    if leave_on_out:
        loo_indices = row_indices
    else:
        loo_indices = [None]

    avg_activity = np.zeros(num_peptides)
    frac_hit = np.zeros(num_peptides)

    for loo_idx in loo_indices:
        subset_indices = np.array([i for i in row_indices if i != loo_idx])
        A_subset = A[subset_indices, :]
        b_subset = b[subset_indices]
        if sparse_solution:
            # L1 minimization to get a small set of confident active peptides
            lasso = Lasso(fit_intercept=False, positive=True)
            lasso.fit(A_subset, b_subset)
            x_with_offset = lasso.coef_
        else:
            # this will work horribly, have fun
            ridge = Ridge(fit_intercept=False, positive=True)
            ridge.fit(A_subset, b_subset)
            x_with_offset = ridge.coef_
        if verbose:
            print("x = %s" % (x,))
            print("c = %s" % (c,))
        x, c = x_with_offset[:-1], x_with_offset[-1]
        avg_activity += x
        frac_hit += (x > min_peptide_activity).astype(float)

    avg_activity /= len(loo_indices)
    frac_hit /= len(loo_indices)
    high_confidence_hits = set(np.where(frac_hit > 0.5)[0])

    return DeconvolutionResult(
        activity_per_peptide=avg_activity,
        prob_hit_per_peptide=frac_hit,
        high_confidence_hits=high_confidence_hits,
    )



from collections import defaultdict
import math
from typing import Optional, Literal

import numpy as np

from .solution import Solution
from .types import PeptidePairList, Replicate, Peptide
from .util import pairs_to_dict, transitive_closure
from .validity import count_violations


def _pools_per_replicate(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    allow_extra_pools: bool,
    verbose: bool,
) -> dict[Replicate, int]:
    too_few_pools = num_peptides < max_peptides_per_pool**2
    if too_few_pools and not allow_extra_pools and verbose:
        print(
            "Warning: cannot satisfy constraints with %d peptides and %d peptides per pool (max allowed %d)"
            % (num_peptides, max_peptides_per_pool, int(np.sqrt(num_peptides)))
        )
    num_pools_first_replicate = int(np.ceil(num_peptides / max_peptides_per_pool))
    result = {0: num_pools_first_replicate}
    for replicate_idx in range(1, num_replicates):
        if too_few_pools and allow_extra_pools:
            max_peptides_per_pool = num_pools_first_replicate
            result[replicate_idx] = int(np.ceil(num_peptides / max_peptides_per_pool))
        else:
            result[replicate_idx] = num_pools_first_replicate
    return result


def _random_peptide_order(
    num_peptides: int, peptide_to_preferred: dict[Peptide, set[Peptide]]
):
    """
    Return the peptide indices in a random order but make sure the ones with preferred neighbors are first
    """
    random_peptide_order = np.arange(num_peptides)
    np.random.shuffle(random_peptide_order)
    assert len(set(random_peptide_order)) == num_peptides
    # assign all peptides with preferred neighbors first
    peptides_with_preferred_neighbors = [
        p for p in random_peptide_order if peptide_to_preferred.get(p)
    ]
    peptides_without_preferred_neighbors = [
        p for p in random_peptide_order if not peptide_to_preferred.get(p)
    ]
    peptide_list = (
        peptides_with_preferred_neighbors + peptides_without_preferred_neighbors
    )
    return peptide_list


def _random_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        peptide_array = np.arange(num_peptides)
        np.random.shuffle(peptide_array)
        pool_assignments = {}
        replicate_to_pool_to_peptides[i] = pool_assignments
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(np.ceil(num_peptides / num_pools))
        for j in range(num_pools):
            start_idx = peptides_per_pool * j
            end_idx = peptides_per_pool * (j + 1)
            pool_assignments[j] = peptide_array[start_idx:end_idx]

    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def _singleton_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    """
    Initialize every peptide to be in its own pool
    """

    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        pool_to_peptides = {}
        for p in range(num_peptides):
            pool_to_peptides[p] = np.array([p])
        replicate_to_pool_to_peptides[i] = pool_to_peptides
    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def _valid_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    peptide_to_invalid = pairs_to_dict(invalid_neighbors)
    peptide_to_preferred = transitive_closure(pairs_to_dict(preferred_neighbors))
    if verbose:
        print("[_valid_init] Invalid neighbors: %s" % (peptide_to_invalid,))
        print("[_valid_init] Preferred neighbors: %s" % (peptide_to_preferred,))

    replicate_to_pool_to_peptides = {}

    for i in range(num_replicates):
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(math.ceil(num_peptides / num_pools))

        peptide_to_pool = {}
        pool_to_peptides = defaultdict(set)

        def add_to_pool(peptide, pool_idx):
            pool = pool_to_peptides[pool_idx]
            pool.add(peptide)
            peptide_to_pool[peptide] = pool_idx
            for other_peptide in pool:
                peptide_to_invalid[peptide].add(other_peptide)
                peptide_to_invalid[other_peptide].add(peptide)

        def curr_num_pools():
            return len(pool_to_peptides)

        def make_new_pool(peptide):
            new_pool_idx = curr_num_pools()
            add_to_pool(peptide, new_pool_idx)
            return new_pool_idx

        for peptide in _random_peptide_order(
            num_peptides=num_peptides, peptide_to_preferred=peptide_to_preferred
        ):
            for preferred_neighbor in peptide_to_preferred.get(peptide, []):
                if preferred_neighbor in peptide_to_invalid[peptide]:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d already invalid with preferred neighbor %d"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool_idx = peptide_to_pool.get(preferred_neighbor)
                if preferred_pool_idx is None:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d, preferred neighbor %d not in a pool yet"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool = pool_to_peptides[preferred_pool_idx]
                if len(preferred_pool) >= peptides_per_pool:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, preferred neighbor %d in pool %d which is already full"
                            % (i, preferred_neighbor, preferred_pool_idx)
                        )
                    continue

                add_to_pool(peptide, preferred_pool_idx)
                if verbose:
                    print(
                        "[_greedy_init] replicate %d, adding peptide %d to preferred pool %d to pair with peptide %d"
                        % (i, peptide, preferred_pool_idx, preferred_neighbor)
                    )
                break
            # if we didn't get a preferred peptide pool that's valid
            # and there's room for more pools, just make a singleton
            if peptide not in peptide_to_pool:
                if curr_num_pools() < num_pools:
                    assigned_pool_idx = make_new_pool(peptide)
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, making new pool %d for peptide %d"
                            % (i, assigned_pool_idx, peptide)
                        )

            # otherwise, try to find a valid pool
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        disallowed_peptides = peptide_to_invalid[peptide]
                        valid = all(
                            [
                                other_peptide not in disallowed_peptides
                                for other_peptide in pool
                            ]
                        )
                        if valid:
                            add_to_pool(peptide, pool_idx)
                            break

            # otherwise, make a new pool
            if peptide not in peptide_to_pool:
                if not allow_extra_pools and verbose:
                    print(
                        "Warning: unable to create valid configuration without exceeding max pools for replicate %d (max %d, adding %d)"
                        % (i + 1, num_pools, curr_num_pools() + 1)
                    )
                make_new_pool(peptide)

        replicate_to_pool_to_peptides[i] = {
            pool_idx: np.array(sorted(peptides))
            for (pool_idx, peptides) in pool_to_peptides.items()
        }
    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def _greedy_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    peptide_to_invalid = pairs_to_dict(invalid_neighbors)
    peptide_to_preferred = transitive_closure(pairs_to_dict(preferred_neighbors))
    if verbose:
        print("[_greedy_init] Invalid neighbors: %s" % (peptide_to_invalid,))
        print("[_greedy_init] Preferred neighbors: %s" % (peptide_to_preferred,))

    replicate_to_pool_to_peptides = {}

    for i in range(num_replicates):
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(math.ceil(num_peptides / num_pools))

        peptide_to_pool = {}
        pool_to_peptides = defaultdict(set)

        def add_to_pool(peptide, pool_idx):
            pool = pool_to_peptides[pool_idx]
            pool.add(peptide)
            peptide_to_pool[peptide] = pool_idx
            for other_peptide in pool:
                peptide_to_invalid[peptide].add(other_peptide)
                peptide_to_invalid[other_peptide].add(peptide)

        def curr_num_pools():
            return len(pool_to_peptides)

        def make_new_pool(peptide):
            new_pool_idx = curr_num_pools()
            add_to_pool(peptide, new_pool_idx)
            return new_pool_idx

        for peptide in _random_peptide_order(
            num_peptides=num_peptides, peptide_to_preferred=peptide_to_preferred
        ):
            for preferred_neighbor in peptide_to_preferred.get(peptide, []):
                if preferred_neighbor in peptide_to_invalid[peptide]:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d already invalid with preferred neighbor %d"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool_idx = peptide_to_pool.get(preferred_neighbor)
                if preferred_pool_idx is None:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d, preferred neighbor %d not in a pool yet"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool = pool_to_peptides[preferred_pool_idx]
                if len(preferred_pool) >= peptides_per_pool:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, preferred neighbor %d in pool %d which is already full"
                            % (i, preferred_neighbor, preferred_pool_idx)
                        )
                    continue

                add_to_pool(peptide, preferred_pool_idx)
                if verbose:
                    print(
                        "[_greedy_init] replicate %d, adding peptide %d to preferred pool %d to pair with peptide %d"
                        % (i, peptide, preferred_pool_idx, preferred_neighbor)
                    )
                break
            # if we didn't get a preferred peptide pool that's valid
            # and there's room for more pools, just make a singleton
            if peptide not in peptide_to_pool:
                if curr_num_pools() < num_pools:
                    assigned_pool_idx = make_new_pool(peptide)
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, making new pool %d for peptide %d"
                            % (i, assigned_pool_idx, peptide)
                        )

            # otherwise, try to find a valid pool
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        disallowed_peptides = peptide_to_invalid[peptide]
                        valid = all(
                            [
                                other_peptide not in disallowed_peptides
                                for other_peptide in pool
                            ]
                        )
                        if valid:
                            add_to_pool(peptide, pool_idx)
                            break

            # otherwise, pick any pool less than the max size
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        add_to_pool(peptide, pool_idx)
                        break

            # lastly, violate the num_pools constraint to make a
            # new singleton anyways
            if peptide not in peptide_to_pool:
                if allow_extra_pools:
                    make_new_pool(peptide)
                else:
                    raise ValueError("Unable to create valid configuration")

        replicate_to_pool_to_peptides[i] = {
            pool_idx: np.array(sorted(peptides))
            for (pool_idx, peptides) in pool_to_peptides.items()
        }

    return Solution(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
    )


def init(
    num_peptides: int = 100,
    max_peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools_per_replicate: Optional[int | dict[Replicate, int]] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    strategy: Literal["greedy", "random", "valid", "singleton"] = "greedy",
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Solution:
    """
    Initialize a Solution for a given configuration

    Args
    ----
    num_peptides
        number of peptides (default: 100)

    peptides_per_pool
        number of peptides per pool (default: 5)

    num_replicates
        number of replicates in the Solution (default: 3)

    num_pools_per_replicate
        number of pools in each replicate (default: ceil(num_peptides / peptides_per_pool))

    invalid_neighbors
        list of peptide pairs that cannot be in the same pool

    preferred_neighbors
        list of peptide pairs that should be in the same pool

    strategy
        initialization strategy, one of {"greedy", "random", "singleton", "valid"} (default: "greedy")

    allow_extra_pools
        whether to allow extra pools to be created to satisfy constraints (default: False)

    verbose
        print diagnostic information during initialization (default: False)
    """
    fn = {
        "greedy": _greedy_init,
        "random": _random_init,
        "singleton": _singleton_init,
        "valid": _valid_init,
    }.get(strategy)
    if num_pools_per_replicate is None:
        num_pools_per_replicate = _pools_per_replicate(
            num_peptides=num_peptides,
            max_peptides_per_pool=max_peptides_per_pool,
            num_replicates=num_replicates,
            allow_extra_pools=allow_extra_pools,
            verbose=verbose,
        )
    elif isinstance(num_pools_per_replicate, int):
        num_pools_per_replicate = {
            i: num_pools_per_replicate for i in range(num_replicates)
        }
    if fn is None:
        raise ValueError("Unknown initialization strategy: '%s'" % (strategy,))

    if verbose:
        print(
            "Using '%s' initialization strategy, pools per replicate: %s"
            % (strategy, num_pools_per_replicate)
        )
    kwargs = dict(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        num_pools_per_replicate=num_pools_per_replicate,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        allow_extra_pools=allow_extra_pools,
        verbose=verbose,
    )
    s = fn(**kwargs)
    violations = count_violations(s)
    if verbose:
        print("Generated solution with %d initial violations" % (violations))
    return s



from golfy import init, optimize, is_valid
from golfy.deconvolution import (
    simulate_elispot_counts,
    create_linear_system,
    solve_linear_system,
)


def test_deconvolution():
    s = init(100, 3, 5)
    optimize(s)
    assert is_valid(s)
    # make sure the margin between background and hit activities
    # is so large that we never falsely identify a background peptide
    counts, hit_peptides = simulate_elispot_counts(
        s,
        num_hits=3,
        max_background_peptide_activity=1,
        min_hit_peptide_activity=100,
    )

    linear_system = create_linear_system(s, counts)
    for loo in [False, True]:
        # solver should work regardless of whether we use LOO to estimate
        # probabilities of hits or not
        solution = solve_linear_system(linear_system, leave_on_out=loo)
        assert solution.high_confidence_hits == hit_peptides



from golfy import find_best_solution


def test_find_best_solution_allow_extra_pools():
    s = find_best_solution(100, 20, 3, allow_extra_pools=True)
    assert s.num_peptides == 100
    assert s.num_replicates == 3
    assert s.max_peptides_per_pool == 20
    assert s.num_pools() > 15


def test_find_best_solution_no_extra_pools():
    s = find_best_solution(100, 20, 3, allow_extra_pools=False)
    assert s.num_peptides == 100
    assert s.num_replicates == 3
    assert s.max_peptides_per_pool == 20
    assert s.num_pools() == 15
